package de.zucizu.ledlights;

import gnu.io.CommPortIdentifier;
import gnu.io.SerialPort;

import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Dimension;
import java.awt.GraphicsConfiguration;
import java.awt.GraphicsDevice;
import java.awt.GraphicsEnvironment;
import java.awt.GridLayout;
import java.awt.Point;
import java.awt.Rectangle;
import java.awt.Robot;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.WindowEvent;
import java.awt.image.BufferedImage;
import java.io.OutputStream;
import java.util.ArrayList;
import java.util.Date;
import java.util.Enumeration;
import java.util.Observable;
import java.util.Observer;

import javax.swing.BorderFactory;
import javax.swing.JButton;
import javax.swing.JCheckBox;
import javax.swing.JComboBox;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JSpinner;
import javax.swing.SpinnerNumberModel;
import javax.swing.event.ChangeEvent;
import javax.swing.event.ChangeListener;

public class AmbientLED implements Observer {

	/**
	 * Time to wait when opening COM port.
	 */
	private static final int SERIAL_TIME_OUT = 1500;

	/**
	 * BPS for communication with Arduinos.
	 */
	private static final int DATA_RATE = 19200;

	/**
	 * Square root of analyzed pixels (steps per for-loop).
	 */
	private int stepWidth = 4;

	/**
	 * User changes this value.
	 */
	private int newStepWidth = 4;

	/**
	 * Time to sleep before analyzing a new screen shot.
	 */
	private int sleeptime = 0;

	/**
	 * Indicates if the GUI shall be updated.
	 */
	private boolean withUIUpdate = true;

	/**
	 * Frame to test functionality.
	 */
	private TestFrame ledTestFrame;

	/**
	 * Frame to set anasysis area.
	 */
	private ConfigFrame ledConfigFrame;

	/**
	 * Indicates, if the testing frame is active.
	 */
	private boolean ledTestFrameActive = false;

	/**
	 * Indicates, whether the configuration frame is active.
	 */
	private boolean ledConfigFrameActive = false;

	/**
	 * Identifier of the port for right and left LEDs.
	 */
	private CommPortIdentifier sidePortID = null;

	/**
	 * Identifier of the port for top LEDs.
	 */
	private CommPortIdentifier topPortID = null;

	/**
	 * This labels change their colors just like the LEDs.
	 */
	private JLabel topLeftLabel;

	/**
	 * @see topLeftLabel
	 */
	private JLabel topRightLabel;

	/**
	 * @see topLeftLabel
	 */
	private JLabel leftLabel;

	/**
	 * @see topLeftLabel
	 */
	private JLabel rightLabel;

	/**
	 * With of the used display.
	 */
	private int screenWidth;

	/**
	 * Height of the used display.
	 */
	private int screenHeight;

	/**
	 * This robot takes screen shots that can be analyzed.
	 */
	private Robot screenshotRobot;

	/**
	 * COM port to communicate with the Arduino controlling the top LEDs.
	 */
	private SerialPort topPort;

	/**
	 * COM port to communicate with the Arduino controlling the side LEDs.
	 */
	private SerialPort sidePort;

	/**
	 * This stream can be used to send commands to the top Arduino.
	 */
	private OutputStream topOutputStream;

	/**
	 * This stream can be used to send commands to the side Arduino.
	 */
	private OutputStream sideOutputStream;

	/**
	 * This thread analyzes the screen shots, calculates mean values and
	 * communicates with the Arduinos.
	 */
	private ScreenshotThread screenshotAnalyzerThread;

	/**
	 * The main frame shows a user interface.
	 */
	private JFrame mainFrame;

	/**
	 * This label shows the time one calculation cycle takes.
	 */
	private JLabel delayLabel;

	/**
	 * The configuration frame needs a reference to this class.
	 */
	private AmbientLED mainApp = this;

	/**
	 * Counts how often the screen shots have been analyzed since last update.
	 */
	private int analyzeCounter = 0;

	/**
	 * Start of analysis.
	 */
	private long analyzeBeginTime = 0;

	/**
	 * End of analysis.
	 */
	private long analyzeEndTime = 0;

	/**
	 * This button starts or stops the testing window.
	 */
	private JButton startTestFrameButton;

	/**
	 * This button starts or stops the configuration frame.
	 */
	private JButton startConfigFrameButton;

	/**
	 * The constructor makes almost everything. It looks for graphic devices,
	 * seeks for available COM ports, starts a new screen shot analyzer thread
	 * and creates a GUI.
	 */
	@SuppressWarnings("unchecked")
	public AmbientLED() {

		// Displays
		GraphicsDevice[] screenDevices = GraphicsEnvironment
				.getLocalGraphicsEnvironment().getScreenDevices();

		// Index of used display.
		int screenIndex = 0;

		if (screenDevices.length > 1) {

			// Dialog if more than one available display.
			ArrayList<String> options = new ArrayList<String>();
			for (GraphicsDevice device : screenDevices) {
				options.add(device.toString());
			}

			Object chosen = JOptionPane
					.showInputDialog(
							null,
							"More than one display available.\nChoose which one to use.",
							"Choose Display", JOptionPane.INFORMATION_MESSAGE,
							null, options.toArray(), options.get(0));
			if (chosen == null) {
				System.exit(0);
			} else {
				screenIndex = options.indexOf(chosen);
			}
		}

		// All calculations need the size of the screen.
		GraphicsConfiguration[] screenConfiguration = screenDevices[screenIndex]
				.getConfigurations();

		screenWidth = screenConfiguration[0].getBounds().width;
		screenHeight = screenConfiguration[0].getBounds().height;

		// The user must be informed that this process may take a little time.
		WaitingFrame waitForComPortsFrame = new WaitingFrame(screenWidth,
				screenHeight, "Looking for COM ports...");
		Thread t = new Thread(waitForComPortsFrame);
		t.start();

		// This robot takes screen shots of the chosen display. If that isn't
		// possible the application doesn't make any sense and can be closed.
		try {
			screenshotRobot = new Robot(screenDevices[screenIndex]);
		} catch (Exception e) {
			waitForComPortsFrame.close();
			JOptionPane.showMessageDialog(null,
					"It is not possible to create a robot on your system."
							+ "\nThat is essential due to the necessity to"
							+ " create screenshots.", "Robot error",
					JOptionPane.ERROR_MESSAGE);
			System.exit(0);
		}

		Enumeration<CommPortIdentifier> comPortIDs = null;

		// Indicates, whether the COM driver is available or not.
		boolean driverAvailable = true;

		// Get all COM Port IDs.
		try {
			comPortIDs = CommPortIdentifier.getPortIdentifiers();
		} catch (UnsatisfiedLinkError err) {
			// Is schown, if the driver is missing.
			JOptionPane
					.showMessageDialog(
							null,
							"It is not possible to load RXTXCommDriver.\n"
									+ "That is necassary to communicate with a microcontroller.\n"
									+ "Make sure you copied 'rxtxSerial.dll' to\n"
									+ System.getProperty("java.home")
									+ System.getProperty("file.separator")
									+ "bin", "Driver Error",
							JOptionPane.ERROR_MESSAGE);
			driverAvailable = false;
		}

		// Names of the ports.
		ArrayList<String> availablePortNamesList = new ArrayList<String>();
		availablePortNamesList.add("null");

		// List of all COM port IDs.
		final ArrayList<CommPortIdentifier> comPortIDsList = new ArrayList<CommPortIdentifier>();
		comPortIDsList.add(null);

		if (driverAvailable) {
			// Adds all available COM ports to the list.
			while (comPortIDs.hasMoreElements()) {
				CommPortIdentifier currPortId = comPortIDs.nextElement();
				if (!currPortId.isCurrentlyOwned()) {
					availablePortNamesList.add(currPortId.getName());
					comPortIDsList.add(currPortId);
				}
			}
		}

		// Array of all ports (their names).
		String[] availablePortNames = new String[availablePortNamesList.size()];

		int i = 0;
		for (String portName : availablePortNamesList) {
			availablePortNames[i++] = portName;
		}

		// Generating user interface.
		topLeftLabel = new JLabel("<html>upper<br></br>left</html>");
		topLeftLabel.setForeground(Color.WHITE);
		topLeftLabel.setOpaque(true);
		topLeftLabel.setBackground(Color.CYAN);
		topRightLabel = new JLabel("<html>upper<br></br>right</html>");
		topRightLabel.setForeground(Color.WHITE);
		topRightLabel.setOpaque(true);
		topRightLabel.setBackground(Color.CYAN);
		rightLabel = new JLabel("right");
		rightLabel.setForeground(Color.WHITE);
		rightLabel.setOpaque(true);
		rightLabel.setBackground(Color.CYAN);
		leftLabel = new JLabel("left_");
		leftLabel.setForeground(Color.WHITE);
		leftLabel.setOpaque(true);
		leftLabel.setBackground(Color.CYAN);

		JLabel comSettingsLabel = new JLabel("<html><u>COM Settings</u></html>");
		JPanel comSettingsPanel = new JPanel();
		comSettingsPanel.add(comSettingsLabel);

		final JCheckBox uiUpdateCheckBox = new JCheckBox();
		uiUpdateCheckBox.setSelected(true);
		uiUpdateCheckBox.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent uiUpdateCheckBoxClickedEvent) {
				JCheckBox uiUpdateBox = (JCheckBox) uiUpdateCheckBoxClickedEvent
						.getSource();
				withUIUpdate = uiUpdateBox.isSelected();
			}
		});

		JPanel uiUpdatePanel = new JPanel();
		uiUpdatePanel.add(new JLabel("GUI Update"));
		uiUpdatePanel.add(uiUpdateCheckBox);

		final JCheckBox comIsActiveCheckBox = new JCheckBox();
		if (!driverAvailable) {
			comIsActiveCheckBox.setEnabled(false);
		}
		comIsActiveCheckBox.addActionListener(new ActionListener() {

			@Override
			public void actionPerformed(ActionEvent comActiveCheckBoxClikedEvent) {
				JCheckBox comActiveBox = (JCheckBox) comActiveCheckBoxClikedEvent
						.getSource();
				if (comActiveBox.isSelected()) {
					if ((sidePortID != null || topPortID != null)
							&& topPortID != sidePortID) {
						System.out.println("COM activated.");
						initSerial();
					} else {
						comActiveBox.setSelected(false);
					}
				} else {
					close();
					System.out.println("COM deactivated.");
				}
			}
		});

		JPanel comIsActivePanel = new JPanel();
		comIsActivePanel.add(new JLabel("COM Active"));
		comIsActivePanel.add(comIsActiveCheckBox);

		JComboBox topPortComboBox = new JComboBox(availablePortNames);
		if (!driverAvailable) {
			topPortComboBox.setEnabled(false);
		}
		JPanel topPortPanel = new JPanel();
		topPortPanel.add(new JLabel("Top"));
		topPortPanel.add(topPortComboBox);
		topPortComboBox.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent portChangedEvent) {
				if (comIsActiveCheckBox.isSelected()) {
					comIsActiveCheckBox.doClick();
				}
				JComboBox topPortBox = (JComboBox) portChangedEvent.getSource();
				topPortID = comPortIDsList.get(topPortBox.getSelectedIndex());
				System.out.print("Top port changed: ");
				if (topPortID != null) {
					System.out.println(topPortID.getName());
				} else {
					System.out.println("null");
				}
			}
		});

		JComboBox sidePortComboBox = new JComboBox(availablePortNames);
		if (!driverAvailable) {
			sidePortComboBox.setEnabled(false);
		}
		JPanel sidePortPanel = new JPanel();
		sidePortPanel.add(new JLabel("Side"));
		sidePortPanel.add(sidePortComboBox);
		sidePortComboBox.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(ActionEvent portChangedEvent) {
				if (comIsActiveCheckBox.isSelected()) {
					comIsActiveCheckBox.doClick();
				}
				JComboBox sidePortBox = (JComboBox) portChangedEvent
						.getSource();
				sidePortID = comPortIDsList.get(sidePortBox.getSelectedIndex());
				System.out.print("Side port changed: ");
				if (sidePortID != null) {
					System.out.println(sidePortID.getName());
				} else {
					System.out.println("null");
				}
			}
		});

		JSpinner stepWidthSpinner = new JSpinner(new SpinnerNumberModel(
				stepWidth, 1, screenHeight < screenWidth ? screenHeight / 10
						: screenWidth / 10, 1));
		stepWidthSpinner.addChangeListener(new ChangeListener() {

			@Override
			public void stateChanged(ChangeEvent stepWidthChangedEvent) {
				newStepWidth = (Integer) ((JSpinner) (stepWidthChangedEvent
						.getSource())).getValue();
				System.out.println("Trying to change step width: "
						+ newStepWidth);
			}
		});

		JPanel stepWidthPanel = new JPanel();
		stepWidthPanel.add(new JLabel("Steps [pixel]"));
		stepWidthPanel.add(stepWidthSpinner);

		JSpinner sleepTimerSpinner = new JSpinner(new SpinnerNumberModel(
				sleeptime, 0, 999, 5));
		sleepTimerSpinner.addChangeListener(new ChangeListener() {

			@Override
			public void stateChanged(ChangeEvent sleepTimeChangedEvent) {
				sleeptime = (Integer) ((JSpinner) (sleepTimeChangedEvent
						.getSource())).getValue();
				System.out.println("Sleep time changed: " + sleeptime);
			}
		});

		JPanel sleepTimePanel = new JPanel();
		sleepTimePanel.add(new JLabel("Sleep [ms]"));
		sleepTimePanel.add(sleepTimerSpinner);

		delayLabel = new JLabel("Update: ");
		JPanel delayPanel = new JPanel();
		delayPanel.add(delayLabel);

		JLabel guiSettingsLabel = new JLabel("<html><u>GUI Settings</u></html>");
		JPanel guiSettingsPanel = new JPanel();
		guiSettingsPanel.add(guiSettingsLabel);

		startTestFrameButton = new JButton("Test");
		startTestFrameButton.addActionListener(new ActionListener() {

			@Override
			public void actionPerformed(ActionEvent e) {
				if (!ledTestFrameActive) {
					startTestFrameButton.setText("Stop test");
					ledTestFrameActive = true;
					ledTestFrame = new TestFrame(mainApp);
					mainFrame.setAlwaysOnTop(true);
				} else {
					startTestFrameButton.setText("Test");
					ledTestFrameActive = false;
					mainFrame.setAlwaysOnTop(false);
					ledTestFrame.close();
				}
			}
		});

		JPanel testframePanel = new JPanel();
		testframePanel.add(startTestFrameButton);

		startConfigFrameButton = new JButton("Config");
		startConfigFrameButton.addActionListener(new ActionListener() {

			@Override
			public void actionPerformed(ActionEvent e) {
				if (!ledConfigFrameActive) {
					startConfigFrameButton.setText("Stop Config");
					ledConfigFrameActive = true;
					ledConfigFrame = new ConfigFrame(mainApp);
				} else {
					startConfigFrameButton.setText("Config");
					ledConfigFrameActive = false;
					ledConfigFrame.close();
				}
			}
		});

		JPanel configFramePanel = new JPanel();
		configFramePanel.add(startConfigFrameButton);

		JPanel spacePanel = new JPanel();

		// If the driver could not be loaded the COM settings will be red.
		Color comColor = driverAvailable ? Color.LIGHT_GRAY : Color.RED;

		JPanel configPanel = new JPanel();
		configPanel.setLayout(new GridLayout(0, 2));
		configPanel.setBorder(BorderFactory.createLineBorder(Color.BLACK, 2));
		comSettingsPanel.setBackground(comColor);
		configPanel.add(comSettingsPanel);
		topPortPanel.setBackground(comColor);
		configPanel.add(topPortPanel);
		comIsActivePanel.setBackground(comColor);
		configPanel.add(comIsActivePanel);
		sidePortPanel.setBackground(comColor);
		configPanel.add(sidePortPanel);
		guiSettingsPanel.setBackground(Color.WHITE);
		configPanel.add(guiSettingsPanel);
		sleepTimePanel.setBackground(Color.WHITE);
		configPanel.add(sleepTimePanel);
		uiUpdatePanel.setBackground(Color.WHITE);
		configPanel.add(uiUpdatePanel);
		stepWidthPanel.setBackground(Color.WHITE);
		configPanel.add(stepWidthPanel);
		delayPanel.setBackground(Color.WHITE);
		configPanel.add(delayPanel);
		spacePanel.setBackground(Color.WHITE);
		configPanel.add(spacePanel);
		configFramePanel.setBackground(Color.WHITE);
		configPanel.add(configFramePanel);
		testframePanel.setBackground(Color.WHITE);
		configPanel.add(testframePanel);

		JPanel topColorPanel = new JPanel(new GridLayout(0, 2));

		topColorPanel.add(topLeftLabel);
		topColorPanel.add(topRightLabel);

		mainFrame = new JFrame("TV LED Light");
		mainFrame.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
		mainFrame.addWindowListener(new java.awt.event.WindowAdapter() {
			public void windowClosing(final WindowEvent winEvt) {
				screenshotAnalyzerThread.cancel();
				close();
				System.exit(0);
			}
		});

		mainFrame.setLayout(new BorderLayout());

		mainFrame.add(topColorPanel, BorderLayout.NORTH);

		mainFrame.add(leftLabel, BorderLayout.WEST);
		mainFrame.add(rightLabel, BorderLayout.EAST);
		mainFrame.add(configPanel, BorderLayout.CENTER);
		mainFrame.pack();

		mainFrame.setLocation((int) (screenWidth - mainFrame.getWidth()) / 2,
				(int) (screenHeight - mainFrame.getHeight()) / 2);

		analyzeBeginTime = new Date().getTime();

		screenshotAnalyzerThread = new ScreenshotThread();
		screenshotAnalyzerThread.addObserver(this);

		waitForComPortsFrame.close();
		mainFrame.setVisible(true);
	}

	/**
	 * Closes all output streams after sending zeros to turn all LEDs off.
	 */
	private void close() {
		if (topOutputStream != null) {
			try {
				topOutputStream.write(0xff);
				for (int z = 0; z < 6; z++) {
					topOutputStream.write(0x00);
				}
				topOutputStream.flush();
			} catch (Exception e) {
				e.printStackTrace();
			}
		}
		if (sideOutputStream != null) {
			try {
				sideOutputStream.write(0xff);
				for (int z = 0; z < 6; z++) {
					sideOutputStream.write(0x00);
				}
				sideOutputStream.flush();
			} catch (Exception e) {
				e.printStackTrace();
			}
		}
		if (topOutputStream != null) {
			try {
				topOutputStream.close();
			} catch (Exception e) {
				System.err.println("Cannot close topOutputStream");
			}
			topOutputStream = null;
		}
		if (topPort != null) {
			try {
				topPort.close();
			} catch (Exception e) {
				System.err.println("Cannot close topLEDsPort");
			}
			topPort = null;
		}
		if (sideOutputStream != null) {
			try {
				sideOutputStream.close();
			} catch (Exception e) {
				System.err.println("Cannot close sideOutputStream");
			}
			sideOutputStream = null;
		}
		if (sidePort != null) {
			try {
				sidePort.close();
			} catch (Exception e) {
				e.printStackTrace();
			}
			sidePort = null;
		}
	}

	/**
	 * Opens the ports and creates output streams to communicate with the
	 * Arduinos.
	 */
	private void initSerial() {
		if (sidePortID != null) {
			try {
				sidePort = (SerialPort) sidePortID.open(this.getClass()
						.getName(), AmbientLED.SERIAL_TIME_OUT);
				sidePort.setSerialPortParams(AmbientLED.DATA_RATE,
						SerialPort.DATABITS_8, SerialPort.STOPBITS_1,
						SerialPort.PARITY_NONE);
				sideOutputStream = sidePort.getOutputStream();
			} catch (Exception e) {
				System.err.println("Cannot open side serial port "
						+ sidePortID.getName() + ".");
			}
		}
		if (topPortID != null) {
			try {
				topPort = (SerialPort) topPortID.open(
						this.getClass().getName(), AmbientLED.SERIAL_TIME_OUT);
				topPort.setSerialPortParams(AmbientLED.DATA_RATE,
						SerialPort.DATABITS_8, SerialPort.STOPBITS_1,
						SerialPort.PARITY_NONE);
				topOutputStream = topPort.getOutputStream();
			} catch (Exception e) {
				System.err.println("Cannot open top serial port "
						+ sidePortID.getName() + ".");
			}
		}
	}

	/**
	 * This method changes the GUI and needs the colors as an array.
	 */
	@Override
	public void update(Observable o, Object colorsToBePainted) {
		int[] colors = (int[]) colorsToBePainted;
		try {
			if (analyzeCounter++ >= 20) {
				analyzeEndTime = new Date().getTime();
				delayLabel.setText("Update: "
						+ ((analyzeEndTime - analyzeBeginTime) / 20) + "ms");
				analyzeBeginTime = analyzeEndTime;
				analyzeCounter = 0;
			}
			topLeftLabel.setBackground(new Color(colors[0], colors[1],
					colors[2]));
			topRightLabel.setBackground(new Color(colors[3], colors[4],
					colors[5]));
			leftLabel.setBackground(new Color(colors[6], colors[7], colors[8]));
			rightLabel.setBackground(new Color(colors[9], colors[10],
					colors[11]));
		} catch (Exception e) {
			System.err.println("Cannot change color in window: ");
			for (int i = 0; i < colors.length; i++) {
				System.err.println("i=" + i + ", Color=" + colors[i]);
			}
		}

	}

	/**
	 * This class takes screen shots, analyzes the colors and sends commands to
	 * the Arduino and the GUI.
	 * 
	 * @author zucizu
	 * 
	 */
	private class ScreenshotThread extends Observable {

		/**
		 * Indicates, that this thread shall stop.
		 */
		private boolean cancel = false;

		/**
		 * Number of all pixels we analyze in the side area.
		 */
		private int numberOfSidePixels = 0;

		/**
		 * Number of all pixels that are inside the top area.
		 */
		private int numberOfTopPixels = 0;

		/**
		 * The X coordinate of the starting point for the top area.
		 */
		private int startXTop = 0;

		/**
		 * The Y coordinate of the starting point for the top area.
		 */
		private int startYTop = 0;

		/**
		 * The X coordinate of the ending point for the top area.
		 */
		private int endXTop = (int) screenWidth / 2;

		/**
		 * The Y coordinate of the ending point for the top area.
		 */
		private int endYTop = (int) screenHeight / 3;

		/**
		 * The X coordinate of the starting point for the side area.
		 */
		private int startXSide = 0;

		/**
		 * The Y coordinate of the starting point for the side area.
		 */
		private int startYSide = 0;

		/**
		 * The X coordinate of the ending point for the side area.
		 */
		private int endXSide = screenWidth / 3;

		/**
		 * The Y coordinate of the ending point for the side area.
		 */
		private int endYSide = screenHeight;

		/**
		 * If the user changed the analysis area the number of piels of course
		 * change too. That's why we have to recalculate that values.
		 */
		private void calculateNewPixelNumbers() {
			int topXLength = endXTop - startXTop;
			int topYLength = endYTop - startYTop;
			numberOfTopPixels = (topXLength * topYLength)
					/ (stepWidth * stepWidth);

			int sideXLength = endXSide - startXSide;
			int sideYLength = endYSide - startYSide;
			numberOfSidePixels = (sideXLength * sideYLength)
					/ (stepWidth * stepWidth);

		}

		/**
		 * This method has to be called, when the analysis areas have changed.
		 * It will recalculate all relevant values.
		 * 
		 * @param points
		 *            The new areas. points[0] = starting point of side area,
		 *            points[1] = ending point of side area, points[2] =
		 *            starting point of top area, points[3] = ending point of
		 *            top area.
		 */
		private void changeAnalysisAreas(Point[] points) {
			this.startXSide = (int) points[0].getX();
			this.startYSide = (int) points[0].getY();

			this.endXSide = (int) points[1].getX();
			this.endYSide = (int) points[1].getY();

			this.startXTop = (int) points[2].getX();
			this.startYTop = (int) points[2].getY();

			this.endXTop = (int) points[3].getX();
			this.endYTop = (int) points[3].getY();

			calculateNewPixelNumbers();
		}

		/**
		 * This constructor creates a new Thread in which all mean values are
		 * calculated. The GUI is being updated and the values are being sent to
		 * the micro controllers.
		 */
		private ScreenshotThread() {

			// Recalculate values if the user changed step width.
			if (newStepWidth != stepWidth) {
				stepWidth = newStepWidth;
				System.out.println("Step width changed: " + stepWidth);
			}

			calculateNewPixelNumbers();

			Thread mainScreenshotThread = new Thread(new Runnable() {

				@Override
				public void run() {
					int[] colors = new int[12];
					try {
						BufferedImage screenShot;
						while (!cancel) {

							// Must be changed here because otherwise
							// calculations would be wrong.
							if (newStepWidth != stepWidth) {
								stepWidth = newStepWidth;
								System.out.println("Step width changed: "
										+ stepWidth);
								calculateNewPixelNumbers();
							}

							int currentPixel;
							float red = 0;
							float green = 0;
							float blue = 0;

							// Take screen shot.
							screenShot = screenshotRobot
									.createScreenCapture(new Rectangle(
											new Dimension(screenWidth,
													screenHeight)));

							int xPosition = 0;
							int yPosition = 0;

							// Calculate mean value of upper left area.
							for (xPosition = startXTop; xPosition < endXTop; xPosition += stepWidth) {
								for (yPosition = startYTop; yPosition < endYTop; yPosition += stepWidth) {
									currentPixel = screenShot.getRGB(xPosition,
											yPosition);
									red += (int) (255 & (currentPixel >> 16));
									green += (int) (255 & (currentPixel >> 8));
									blue += (int) (255 & (currentPixel));
								}
							}

							red /= numberOfTopPixels;
							green /= numberOfTopPixels;
							blue /= numberOfTopPixels;

							// save the values.
							colors[0] = (int) red;
							colors[1] = (int) green;
							colors[2] = (int) blue;

							red = green = blue = 0;

							// Upper right
							for (xPosition = screenWidth - endXTop; xPosition < screenWidth
									- startXTop; xPosition += stepWidth) {
								for (yPosition = startYTop; yPosition < endYTop; yPosition += stepWidth) {
									currentPixel = screenShot.getRGB(xPosition,
											yPosition);
									red += (int) (255 & (currentPixel >> 16));
									green += (int) (255 & (currentPixel >> 8));
									blue += (int) (255 & (currentPixel));
								}
							}

							red /= numberOfTopPixels;
							green /= numberOfTopPixels;
							blue /= numberOfTopPixels;

							colors[3] = (int) red;
							colors[4] = (int) green;
							colors[5] = (int) blue;

							red = green = blue = 0;

							// Left
							for (xPosition = startXSide; xPosition < endXSide; xPosition += stepWidth) {
								for (yPosition = startYSide; yPosition < endYSide; yPosition += stepWidth) {
									currentPixel = screenShot.getRGB(xPosition,
											yPosition);
									red += (int) (255 & (currentPixel >> 16));
									green += (int) (255 & (currentPixel >> 8));
									blue += (int) (255 & (currentPixel));
								}
							}

							red /= numberOfSidePixels;
							green /= numberOfSidePixels;
							blue /= numberOfSidePixels;

							colors[6] = (int) red;
							colors[7] = (int) green;
							colors[8] = (int) blue;

							red = green = blue = 0;

							// Right
							for (xPosition = screenWidth - endXSide; xPosition < screenWidth
									- startXSide; xPosition += stepWidth) {
								for (yPosition = startYSide; yPosition < endYSide; yPosition += stepWidth) {
									currentPixel = screenShot.getRGB(xPosition,
											yPosition);
									red += (int) (255 & (currentPixel >> 16));
									green += (int) (255 & (currentPixel >> 8));
									blue += (int) (255 & (currentPixel));
								}
							}

							red /= numberOfSidePixels;
							green /= numberOfSidePixels;
							blue /= numberOfSidePixels;

							colors[9] = (int) red;
							colors[10] = (int) green;
							colors[11] = (int) blue;

							for (int i = 0; i < colors.length; i++) {
								if (colors[i] > 255) {
									colors[i] = 255;
								}
							}

							// Probably inform the GUI.
							if (withUIUpdate) {
								setChanged();
								notifyObservers(colors);
								clearChanged();
							}

							// Send values to the top micro controller.
							if (topOutputStream != null) {
								try {
									topOutputStream.write(0xff);
									topOutputStream.write((byte) colors[0]);
									topOutputStream.write((byte) colors[1]);
									topOutputStream.write((byte) colors[2]);
									topOutputStream.write((byte) colors[3]);
									topOutputStream.write((byte) colors[4]);
									topOutputStream.write((byte) colors[5]);
									topOutputStream.flush();
								} catch (Exception e) {
									System.err
											.println("Problem writing to top output stream. Maybe it's already closed.");
								}
							}

							// Send values to the side micro controller.
							if (sideOutputStream != null) {
								try {
									sideOutputStream.write(0xff);
									sideOutputStream.write((byte) colors[6]);
									sideOutputStream.write((byte) colors[7]);
									sideOutputStream.write((byte) colors[8]);
									sideOutputStream.write((byte) colors[9]);
									sideOutputStream.write((byte) colors[10]);
									sideOutputStream.write((byte) colors[11]);
									sideOutputStream.flush();
								} catch (Exception e) {
									System.err
											.println("Problem writing to side output stream. Maybe it's already closed.");
								}
							}
							Thread.sleep(sleeptime);
						}
					} catch (Exception e) {
						e.printStackTrace();
					}
				}
			});
			mainScreenshotThread.start();
		}

		/**
		 * Cancels this thread.
		 */
		private void cancel() {
			this.cancel = true;
		}
	}

	/**
	 * Must be called when the testing frame is closed from somewhere else.
	 */
	public void testFrameClosed() {
		startTestFrameButton.setText("Test");
		ledTestFrameActive = false;
		mainFrame.setAlwaysOnTop(false);
	}

	/**
	 * Must be called when the configuration frame is closed from somewhere
	 * else.
	 */
	public void configFrameClosed(Point[] points) {
		startConfigFrameButton.setText("Config");
		ledConfigFrameActive = false;
		screenshotAnalyzerThread.changeAnalysisAreas(points);
	}

	/**
	 * Creates an image of the chosen display.
	 * 
	 * @return A screen shot of the whole screen without the main frame.
	 */
	public BufferedImage getScreenShot() {
		// We don't want our window to be in that screenshot.
		mainFrame.setVisible(false);
		// Window needs some time to close.
		try {
			Thread.sleep(300);
		} catch (InterruptedException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}
		BufferedImage screen = screenshotRobot
				.createScreenCapture(new Rectangle(new Dimension(screenWidth,
						screenHeight)));
		mainFrame.setVisible(true);
		return screen;
	}

	/**
	 * Starts a new instance of this class.
	 * 
	 * @param args
	 *            No arguments needed.
	 */
	public static void main(String[] args) {
		new AmbientLED();
	}
}